@@ -1,5 +1,6 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TupleSections #-}
 
 module Network.Wai.Handler.Warp.Response (
     sendResponse
@@ -16,64 +17,92 @@ import qualified Data.CaseInsensitive as CI
 import Data.Conduit
 import Data.Conduit.Blaze (builderToByteString)
 import qualified Data.Conduit.List as CL
-import Data.Maybe (isJust)
+import Data.Maybe (isJust, listToMaybe)
 import Data.Monoid (mappend)
 import qualified Network.HTTP.Types as H
 import Network.Wai
 import Network.Wai.Handler.Warp.ReadInt
+import Network.Wai.Handler.Warp.Settings
 import qualified Network.Wai.Handler.Warp.ResponseHeader as RH
 import qualified Network.Wai.Handler.Warp.Timeout as T
 import Network.Wai.Handler.Warp.Types
 import qualified System.PosixCompat.Files as P
+import Network.HTTP.Attoparsec (parseByteRanges)
+import Numeric (showInt)
 
 ----------------------------------------------------------------
 ----------------------------------------------------------------
 
-sendResponse :: Cleaner -> Request -> Connection -> Response
+sendResponse :: Settings
+             -> Cleaner -> Request -> Connection -> Response
              -> ResourceT IO Bool
 
 ----------------------------------------------------------------
 
-sendResponse cleaner req conn (ResponseFile s hs path mpart) =
+sendResponse settings cleaner req conn (ResponseFile s0 hs0 path mpart0) =
     headerAndLength >>= sendResponse'
   where
+    hs = addAccept hs0
     th = threadHandle cleaner
-    headerAndLength = case (readInt <$> checkLength hs, mpart) of
-        (Just cl, _)         -> return $ Right (hs, cl)
-        (Nothing, Nothing)   -> liftIO . try $ do
-            cl <- fromIntegral . P.fileSize <$> P.getFileStatus path
-            return (addLength cl hs, cl)
-        (Nothing, Just part) -> do
-            let cl = fromIntegral $ filePartByteCount part
-            return $ Right (addLength cl hs, cl)
+    headerAndLength = liftIO . try $ do
+        (hadLength, cl) <-
+            case readInt <$> checkLength hs of
+                Just cl -> return (True, cl)
+                Nothing ->
+                    case mpart0 of
+                        Just part -> return (False, fromIntegral $ filePartByteCount part)
+                        Nothing -> (False, ) . fromIntegral . P.fileSize
+                               <$> P.getFileStatus path
+        let (s, addRange, beg, end) =
+                case mpart0 of
+                    Just part -> (s0, id, filePartOffset part, filePartByteCount part)
+                    Nothing ->
+                        case lookup H.hRange (requestHeaders req) >>= parseByteRanges >>= listToMaybe of
+                            Just range | s0 == H.status200 ->
+                                case range of
+                                    H.ByteRangeFrom from -> rangeRes cl from (cl - 1)
+                                    H.ByteRangeFromTo from to -> rangeRes cl from to
+                                    H.ByteRangeSuffix count -> rangeRes cl (cl - count) (cl - 1)
+                            _ -> (s0, id, 0, cl)
+            hs'
+                | hadLength = hs
+                | otherwise = addLength end hs
+        return (s, addRange hs', beg, end)
 
-    sendResponse' (Right (lengthyHeaders, cl))
+    rangeRes cl from to = (H.status206, (("Content-Range", rangeHeader cl from to):), from, to - from + 1)
+
+    rangeHeader total from to = B.pack
+      $ 'b' : 'y': 't' : 'e' : 's' : ' '
+      : showInt from
+      ( '-'
+      : showInt to
+      ( '/'
+      : showInt total ""))
+
+    sendResponse' (Right (s, lengthyHeaders, beg, end))
       | hasBody s req = liftIO $ do
-          lheader <- composeHeader version s lengthyHeaders
+          lheader <- composeHeader settings version s lengthyHeaders
           connSendFile conn path beg end (T.tickle th) [lheader] cleaner
           T.tickle th
           return isPersist
       | otherwise = liftIO $ do
-          composeHeader version s hs >>= connSendAll conn
+          composeHeader settings version s hs >>= connSendAll conn
           T.tickle th
           return isPersist -- FIXME isKeepAlive?
       where
-        (beg,end) = case mpart of
-            Nothing  -> (0,cl)
-            Just prt -> (filePartOffset prt, filePartByteCount prt)
         version = httpVersion req
         (isPersist,_) = infoFromRequest req
 
     sendResponse' (Left (_ :: SomeException)) =
-        sendResponse cleaner req conn notFound
+        sendResponse settings cleaner req conn notFound
       where
         notFound = responseLBS H.status404 [(H.hContentType, "text/plain")] "File not found"
 
 ----------------------------------------------------------------
 
-sendResponse cleaner req conn (ResponseBuilder s hs b)
+sendResponse settings cleaner req conn (ResponseBuilder s hs b)
   | hasBody s req = liftIO $ do
-      header <- composeHeaderBuilder version s hs needsChunked
+      header <- composeHeaderBuilder settings version s hs needsChunked
       let body
             | needsChunked = header `mappend` chunkedTransferEncoding b
                                     `mappend` chunkedTransferTerminator
@@ -83,7 +112,7 @@ sendResponse cleaner req conn (ResponseBuilder s hs b)
           T.tickle th
       return isKeepAlive
   | otherwise = liftIO $ do
-      composeHeader version s hs >>= connSendAll conn
+      composeHeader settings version s hs >>= connSendAll conn
       T.tickle th
       return isPersist
   where
@@ -94,14 +123,14 @@ sendResponse cleaner req conn (ResponseBuilder s hs b)
 
 ----------------------------------------------------------------
 
-sendResponse cleaner req conn (ResponseSource s hs bodyFlush)
+sendResponse settings cleaner req conn (ResponseSource s hs bodyFlush)
   | hasBody s req = do
-      header <- liftIO $ composeHeaderBuilder version s hs needsChunked
+      header <- liftIO $ composeHeaderBuilder settings version s hs needsChunked
       let src = CL.sourceList [header] `mappend` cbody
       src $$ builderToByteString =$ connSink conn th
       return isKeepAlive
   | otherwise = liftIO $ do
-      composeHeader version s hs >>= connSendAll conn
+      composeHeader settings version s hs >>= connSendAll conn
       T.tickle th
       return isPersist
   where
@@ -190,26 +219,33 @@ hasBody s req = sc /= 204
 addLength :: Integer -> H.ResponseHeaders -> H.ResponseHeaders
 addLength cl hdrs = (H.hContentLength, B.pack $ show cl) : hdrs
 
+addAccept :: H.ResponseHeaders -> H.ResponseHeaders
+addAccept = (("Accept-Ranges", "bytes"):)
+
 addEncodingHeader :: H.ResponseHeaders -> H.ResponseHeaders
 addEncodingHeader hdrs = (hTransferEncoding, "chunked") : hdrs
 
-addServerHeader :: H.ResponseHeaders -> H.ResponseHeaders
-addServerHeader hdrs = case lookup hServer hdrs of
-    Nothing -> warpVersionHeader : hdrs
+addServerHeader :: Settings -> H.ResponseHeaders -> H.ResponseHeaders
+addServerHeader settings hdrs = case lookup hServer hdrs of
+    Nothing -> warpVersionHeader settings : hdrs
     Just _  -> hdrs
 
-warpVersionHeader :: H.Header
-warpVersionHeader = (hServer, ver)
-  where
-    ver = B.pack $ "Warp/" ++ warpVersion
+warpVersionHeader :: Settings -> H.Header
+warpVersionHeader settings = (hServer, settingsServerName settings)
 
 ----------------------------------------------------------------
 
-composeHeader :: H.HttpVersion -> H.Status -> H.ResponseHeaders -> IO ByteString
-composeHeader version s hs = RH.composeHeader version s (addServerHeader hs)
+composeHeader :: Settings
+              -> H.HttpVersion
+              -> H.Status
+              -> H.ResponseHeaders
+              -> IO ByteString
+composeHeader settings version s hs =
+    RH.composeHeader version s
+  $ addServerHeader settings hs
 
-composeHeaderBuilder :: H.HttpVersion -> H.Status -> H.ResponseHeaders -> Bool -> IO Builder
-composeHeaderBuilder ver s hs True =
-    fromByteString <$> composeHeader ver s (addEncodingHeader hs)
-composeHeaderBuilder ver s hs False =
-    fromByteString <$> composeHeader ver s hs
+composeHeaderBuilder :: Settings -> H.HttpVersion -> H.Status -> H.ResponseHeaders -> Bool -> IO Builder
+composeHeaderBuilder settings ver s hs True =
+    fromByteString <$> composeHeader settings ver s (addEncodingHeader hs)
+composeHeaderBuilder settings ver s hs False =
+    fromByteString <$> composeHeader settings ver s hs
