@@ -1,14 +1,16 @@
 module Network.Wai.Handler.Warp.Settings where
 
 import Control.Exception
-import Control.Monad
 import qualified Data.ByteString as S
+import qualified Data.ByteString.Char8 as S8
 import Data.Conduit
 import Data.Conduit.Network (HostPreference (HostIPv4))
+import GHC.IO.Exception (IOErrorType(..))
 import Network.Wai
 import Network.Wai.Handler.Warp.Timeout
 import Network.Wai.Handler.Warp.Types
 import System.IO (hPrint, stderr)
+import System.IO.Error (ioeGetErrorType)
 
 -- | Various Warp server settings. This is purposely kept as an abstract data
 -- type so that new settings can be added without breaking backwards
@@ -39,6 +41,12 @@ data Settings = Settings
       -- Default: do nothing.
       --
       -- Since 1.3.6
+    , settingsServerName :: S.ByteString
+      -- ^ Server name to be sent in the Server header.
+      --
+      -- Default: Warp\//version/
+      --
+      -- Since 1.3.8
     }
 
 -- | The default settings for the Warp server. See the individual settings for
@@ -47,12 +55,7 @@ defaultSettings :: Settings
 defaultSettings = Settings
     { settingsPort = 3000
     , settingsHost = HostIPv4
-    , settingsOnException = \e ->
-        case fromException e of
-            Just x -> go x
-            Nothing ->
-                when (go' $ fromException e) $
-                    hPrint stderr e
+    , settingsOnException = defaultExceptionHandler
     , settingsOnOpen = return ()
     , settingsOnClose = return ()
     , settingsTimeout = 30
@@ -61,9 +64,27 @@ defaultSettings = Settings
     , settingsFdCacheDuration = 10
     , settingsResourceTPerRequest = True
     , settingsBeforeMainLoop = return ()
+    , settingsServerName = S8.pack $ "Warp/" ++ warpVersion
     }
+
+defaultExceptionHandler :: SomeException -> IO ()
+defaultExceptionHandler e = throwIO e `catches` handlers
   where
-    go :: InvalidRequest -> IO ()
-    go _ = return ()
-    go' (Just ThreadKilled) = False
-    go' _ = True
+    handlers = [Handler ah, Handler ih, Handler oh, Handler sh]
+
+    ah :: AsyncException -> IO ()
+    ah ThreadKilled = return ()
+    ah x            = hPrint stderr x
+
+    ih :: InvalidRequest -> IO ()
+    ih _ = return ()
+
+    oh :: IOException -> IO ()
+    oh x
+      | et == ResourceVanished || et == InvalidArgument = return ()
+      | otherwise         = hPrint stderr x
+      where
+        et = ioeGetErrorType x
+
+    sh :: SomeException -> IO ()
+    sh x = hPrint stderr x
